#!/usr/bin/env python
from __future__ import absolute_import, division, print_function, with_statement

"""
task: wraps a GMTK subtask to reduce size of output
"""

# Copyright 2009-2013 Michael M. Hoffman <michael.hoffman@utoronto.ca>

from ast import literal_eval
from contextlib import contextmanager
from errno import ENOENT
import gc
from os import extsep, fdopen, EX_TEMPFAIL, remove
import re
import sys
from tempfile import mkstemp

from numpy import argmax, array, empty, where, diff, r_, zeros
import optbuild
from six.moves import map, range, zip

from .observations import (make_continuous_cells, make_supervision_cells,
                           make_virtual_evidence_cells, _save_window)
from ._util import (BED_SCORE, BED_STRAND, ceildiv, DTYPE_IDENTIFY, EXT_FLOAT,
                    EXT_INT, EXT_LIST, EXT_VIRTUAL_EVIDENCE, extract_superlabel,
                    fill_array, find_segment_starts, get_label_color, TRAIN_PROG,
                    POSTERIOR_PROG, POSTERIOR_SCALE_FACTOR, read_posterior,
                    SEGWAY_ENCODING, VALIDATE_PROG, VITERBI_PROG,
                    VIRTUAL_EVIDENCE_LIST_FILENAME_PLACEHOLDER)

MSG_SUCCESS = "____ PROGRAM ENDED SUCCESSFULLY WITH STATUS 0 AT"

SCORE_MIN = 100
SCORE_MAX = 1000

SEG_INVALID = -1

EXT_OPTIONS = {}
EXT_OPTIONS[EXT_FLOAT] = "-of1"  # duplicative of run.py
EXT_OPTIONS[EXT_INT] = "-of2"

USAGE = "args: VERB KIND OUTFILE CHROM START END RESOLUTION REVERSE [ARGS...]"

# Dummy observation filename required for gmtk EM bundling. Has no effect but a
# name is necessary.
PLACEHOLDER_OBSERVATION_FILENAME = "/dev/zero"

GMTK_TRRNG_OPTION_STRING = "-trrng"  # Range to train over segment file


@contextmanager
def mkstemp_observation(chromosome_name, start, end, suffix):
    """A context manager that provides a tuple of a file object and full
    filepath of an a observation file with the given suffix.
    Does not delete the created temporary file after exiting."""

    # Set a common prefix for the observation files
    prefix = "{}.{}.{}.".format(chromosome_name, start, end)
    fd, filename = mkstemp(suffix=extsep + suffix, prefix=prefix)

    temp_observation_file = fdopen(fd, "w")

    yield temp_observation_file, filename

    temp_observation_file.close()


def save_temp_observations(chromosome_name, start, end, continuous_cells,
                           resolution, distribution, supervision_data,
                           virtual_evidence_data, num_labels):
    """Returns a tuple (float_obs, int_obs) of temporary filepaths for the
    int/float observation filenames unique to this process"""

    # Create secure temporary observation files
    with mkstemp_observation(chromosome_name, start, end, EXT_FLOAT) as \
            (float_observations_file, float_observations_filename), \
            mkstemp_observation(chromosome_name, start, end, EXT_INT) as \
            (int_observations_file, int_observations_filename), \
            mkstemp_observation(chromosome_name, start, end, EXT_VIRTUAL_EVIDENCE) as \
            (virtual_evidence_file, virtual_evidence_filename):

            # numpy's tofile (which is used) can take an open python file
            # object
            # XXX: Currently seq_data is disabled until dinucleotide is enabled
            _save_window(float_observations_file, int_observations_file,
                         continuous_cells, resolution, distribution,
                         seq_data=None, supervision_data=supervision_data,
                         virtual_evidence_data=virtual_evidence_data,
                         virtual_evidence_filename_or_file=virtual_evidence_file,
                         num_labels=num_labels)

    return float_observations_filename, int_observations_filename, \
           virtual_evidence_filename


def save_temp_observation_filelists(float_observations_filename,
                                    int_observations_filename,
                                    virtual_evidence_observations_filename):
    """Create an observation file list containing the respective observation
    file name.

    Returns a tuple (float_obs_list, int_obs_list) of temporary filepaths
    for the int/float observation lists (files) unique to this process
    """

    # Create secure temporary observation files
    float_observation_list_fd, float_observation_list_filename = \
        mkstemp(prefix=EXT_FLOAT + extsep, suffix=extsep + EXT_LIST)
    int_observation_list_fd, int_observation_list_filename = \
        mkstemp(prefix=EXT_INT + extsep, suffix=extsep + EXT_LIST)
    virtual_evidence_observation_list_fd,\
        virtual_evidence_observation_list_filename = \
        mkstemp(prefix=EXT_VIRTUAL_EVIDENCE + extsep, suffix=extsep + EXT_LIST)
    # Write out the observation filename to their respective observation list
    # For gmtk observation list files, there may be more than one
    # observation file. In this case we only ever insert one
    # print_to_fd uses a context manager which implicity closes the
    # os-level file descriptor
    print_to_fd(float_observation_list_fd, float_observations_filename)
    print_to_fd(int_observation_list_fd, int_observations_filename)
    print_to_fd(virtual_evidence_observation_list_fd,
                virtual_evidence_observations_filename)

    return float_observation_list_filename, int_observation_list_filename, \
           virtual_evidence_observation_list_filename


def replace_subsequent_value(input_list, query, new):
    """Attempts to modify the given input list with no exception so that the
    value following the query is modified to the new value """

    try:
        new_index = input_list.index(query) + 1
        input_list[new_index] = new
    # If the query value is not found
    except ValueError:
        # Do nothing
        pass
    # If the new index is out of range
    except IndexError:
        # Do nothing
        pass

def prepare_virtual_evidence(virtual_evidence, start, end, num_labels,
                             virtual_evidence_coords, virtual_evidence_priors):
    if virtual_evidence == "False":
        return None
    virtual_evidence_coords = literal_eval(virtual_evidence_coords)
    virtual_evidence_priors = literal_eval(virtual_evidence_priors)

    virtual_evidence_cells = make_virtual_evidence_cells(
               virtual_evidence_coords,
               virtual_evidence_priors,
               start, end, num_labels)

    return virtual_evidence_cells

def prepare_gmtk_observations(gmtk_args, chromosome_name, start, end,
                              continuous_cells, resolution, distribution,
                              supervision_data=None, virtual_evidence_data=None,
                              num_labels=None):
    """Returns a list of filepaths to observation files created for gmtk
    and modifies the necessary arguments (args) for running gmtk"""

    try:
        # Create the gmtk observation files
        float_observations_filename, int_observations_filename, \
            virtual_evidence_filename = \
            save_temp_observations(chromosome_name, start, end,
                                   continuous_cells, resolution, distribution,
                                   supervision_data, virtual_evidence_data,
                                   num_labels)

        # Create the gmtk observation file lists
        float_observation_list_filename, int_observation_list_filename, \
            virtual_evidence_list_filename = \
            save_temp_observation_filelists(float_observations_filename,
                                            int_observations_filename,
                                            virtual_evidence_filename)
    # If any exception occurred
    except:  # NOQA
        # Attempt to remove any created files
        force_remove_all_files([float_observations_filename,
                                int_observations_filename,
                                virtual_evidence_filename,
                                float_observation_list_filename,
                                int_observation_list_filename,
                                virtual_evidence_list_filename])
        # Reraise the exception
        raise

    # Modify the given gmtk arguments to use the temporary observation lists
    replace_subsequent_value(gmtk_args, EXT_OPTIONS[EXT_FLOAT],
                             float_observation_list_filename)
    replace_subsequent_value(gmtk_args, EXT_OPTIONS[EXT_INT],
                             int_observation_list_filename)

    # cppCommandOptions is stored as a string with format CPP_DIRECTIVE_FMT
    cpp_command_options_index = gmtk_args.index("-cppCommandOptions") + 1
    cpp_command_str = gmtk_args[cpp_command_options_index]
    # if the placeholder is present, it is replaced. otherwise, the cpp options
    # are unchanged
    gmtk_args[cpp_command_options_index] = cpp_command_str.replace(
                                    VIRTUAL_EVIDENCE_LIST_FILENAME_PLACEHOLDER,
                                    virtual_evidence_list_filename, 1)

    # Modify the given gmtk arguments so only the first (and only) file in the
    # observation lists are used
    replace_subsequent_value(gmtk_args, GMTK_TRRNG_OPTION_STRING, "0")

    # Return the list of filenames created
    return [float_observations_filename, int_observations_filename,
            virtual_evidence_filename,
            float_observation_list_filename, int_observation_list_filename,
            virtual_evidence_list_filename]


@contextmanager
def files_to_remove(filenames):
    """Creates a context manager where upon exit, ensures files are removed"""
    yield

    force_remove_all_files(filenames)


def force_remove_all_files(filenames):
    exception_list = []
    # For each file name
    for filename in filenames:
        # Attempt to remove the file
        try:
            force_remove_file(filename)
        # Catch any exception
        except:  # NOQA
            # Store the exception
            # The 2nd element from exc_info is the exception object instance
            exception_list.append(sys.exc_info()[1])

    handle_multiple_exceptions(exception_list)


def force_remove_file(filename):
    """Attempts to remove the given filename. Catches and ignores an exception
    if the file does not exist and raises all other exceptions"""
    try:
        remove(filename)
    # Ignore exceptions where the file does not exist
    except OSError as err:
        # If a different exception was found
        if err.errno != ENOENT:
            # Reraise
            raise


def handle_multiple_exceptions(exception_list):
    """Takes an list of exception objects and prints out all information to
    stderr and raises the first exception (if any exceptions exist)."""
    # If any exceptions exist
    if exception_list:
        # Print out all exceptions and raise the first
        first_exception = exception_list.pop(0)
        for exception in exception_list:
            print(exception, file=sys.stderr)
        raise first_exception


def make_track_indexes(text):
    return array(text.split(","), int)


def divide_posterior_array(posterior_code, num_frames, num_sublabels):
    """
    takes a one-dimensional array whose values are integers of the form
    label * num_sublabels + sublabel
    and creates a two-dimensional array whose columns contain the label
    and the sublabel in separate values. This is a convenience function to
    provide the find_segment_starts() function with data in the same format
    as during the viterbi task.
    """
    res = zeros((2, num_frames), DTYPE_IDENTIFY)
    for frame_index in range(num_frames):
        total_label = posterior_code[frame_index]
        label, sublabel = divmod(total_label, num_sublabels)
        res[:, frame_index] = array([label, sublabel])
    return res


def parse_viterbi(lines, do_reverse=False, output_label="seg"):
    """
    returns: numpy.ndarray of size (num_frames,), type DTYPE_IDENTIFY
    """
    lines = iter(lines)

    # Segment 0, after Island[...]
    assert next(lines).startswith("Segment ")

    # ========
    assert next(lines).startswith("========")

    # Segment 0, number of frames = 1001, viteri-score = -6998.363710
    line = next(lines)
    assert line.startswith("Segment ")

    num_frames_text = line.split(", ")[1].partition(" = ")
    assert num_frames_text[0] == "number of frames"
    assert num_frames_text[1] == " = "

    num_frames = int(num_frames_text[2])

    # Printing random variables from (P,C,E)=(1,999,0) partitions
    line = next(lines)
    assert line.startswith("Printing random variables from")
    seg_dict = {'seg': 0, 'subseg': 1}
    # if output_label == "subseg" or "full", need to catch
    # subseg output
    if output_label != "seg":
        re_seg = re.compile(r"^(seg|subseg)\((\d+)\)=(\d+)$")
    else:
        re_seg = re.compile(r"^(seg)\((\d+)\)=(\d+)$")
    # sentinel value
    res = fill_array(SEG_INVALID, (2, num_frames), DTYPE_IDENTIFY)
    for line in lines:
        # Ptn-0 P': seg(0)=24,seg(1)=24
        if line.startswith(MSG_SUCCESS):
            assert (res[0] != SEG_INVALID).all()
            # if output_label == "subseg" or "full",
            # res will have 2 rows
            if output_label != "seg":
                assert (res[1] != SEG_INVALID).all()
                return res
            else:
                return res[0]

        assert line.startswith("Ptn-")

        values = line.rpartition(": ")[2]

        for pair in values.split(","):
            match = re_seg.match(pair)
            if not match:
                continue

            index = int(match.group(2))
            if do_reverse:
                index = -1 - index  # -1, -2, -3, etc.

            val = int(match.group(3))
            seg_index = seg_dict[match.group(1)]
            res[seg_index][index] = val

    # shouldn't get to this point
    raise ValueError("%s did not complete successfully" % VITERBI_PROG.prog)


# num_cols is for use by genomedata_count_condition
# XXX: should move this function somewhere else

def write_bed(outfile, start_pos, labels, coord, resolution, num_labels,
              num_cols=None, num_sublabels=None, sublabels=None):
    """
    start_pos is an array
    """
    (chrom, region_start, region_end) = coord

    start_pos = region_start + (start_pos * resolution)

    # correct last position which may not be a full sample beyond
    if __debug__:
        remainder = (region_end - region_start) % resolution
        if remainder == 0:
            remainder = resolution
        assert region_end == start_pos[-1] - resolution + remainder
    start_pos[-1] = region_end

    # score_step = (SCORE_MAX - SCORE_MIN) / (num_labels - 1)
    label_colors = [get_label_color(extract_superlabel(seg_label)) for
                    seg_label in labels]

    zipper = zip(start_pos[:-1], start_pos[1:], labels, label_colors)

    # this is easily concatenated since it has no context
    for seg_start, seg_end, seg_label, label_color in zipper:
        name = str(seg_label)

        chrom_start = str(seg_start)
        chrom_end = str(seg_end)
        item_rgb = label_color

        row = [chrom, chrom_start, chrom_end, name, BED_SCORE, BED_STRAND,
               chrom_start, chrom_end, item_rgb][:num_cols]

        print(*row, sep="\t", file=outfile)

    # assert that the whole region is mapped
    # seg_end here means the last seg_end in the loop
    assert seg_end == region_end


def save_bed(outfilename, *args, **kwargs):
    with open(outfilename, "w") as outfile:
        write_bed(outfile, *args, **kwargs)


def read_posterior_save_bed(coord, resolution, do_reverse,
                            outfilename_tmpl, num_labels, infile,
                            num_sublabels, output_label):
    if do_reverse:
        raise NotImplementedError
    num_sublabels = int(num_sublabels)
    (chrom, start, end) = coord
    num_frames = ceildiv(end - start, resolution)
    probs = read_posterior(infile, num_frames, num_labels,
                           num_sublabels, output_label)
    probs_rounded = empty(probs.shape, float)  # casted to int after rounding

    # Write posterior code file
    posterior_code = argmax(probs, axis=1)
    if output_label != "seg":
        posterior_code = divide_posterior_array(posterior_code, num_frames,
                                                num_sublabels)
    start_pos, labels = find_segment_starts(posterior_code, output_label)
    bed_filename = outfilename_tmpl % "_code"
    save_bed(bed_filename, start_pos, labels, coord, resolution,
             int(num_labels))
    if output_label == "subseg":
        label_print_range = range(num_labels * num_sublabels)
        label_names = label_print_range
    elif output_label == "full":
        label_print_range = range(num_labels * num_sublabels)
        label_names = ("%d.%d" % divmod(label, num_sublabels)
                       for label in label_print_range)
    else:
        label_print_range = range(num_labels)
        label_names = label_print_range

    # Write label-wise posterior bedgraph files
    outfilenames = []
    for label_index in label_names:
        outfilenames.append(outfilename_tmpl % label_index)

    # scale, round, and cast to int
    (probs * POSTERIOR_SCALE_FACTOR).round(out=probs_rounded)
    probs_rounded = probs_rounded.astype(int)

    # print array columns as text to each outfile
    zipper = zip(outfilenames, probs_rounded.T, label_print_range)
    for outfilename, probs_rounded_label, label_index in zipper:
        # run-length encoding on the probs_rounded_label

        with open(outfilename, "w") as outfile:
            # Create a list of indicies of unique values in the probability
            # BED labels
            unique_prob_value_indices, = where(diff(probs_rounded_label) != 0)
            # The first index is always unique
            unique_prob_value_indices = r_[0, unique_prob_value_indices + 1]
            region_coords = r_[(unique_prob_value_indices * resolution) +
                               start,
                               end]

            bed_zipper = zip(region_coords[:-1], region_coords[1:],
                             unique_prob_value_indices)

            for bed_start, bed_end, prob_index in bed_zipper:
                chrom_start = str(bed_start)
                chrom_end = str(bed_end)
                value = str(probs_rounded_label[prob_index])

                row = [chrom, chrom_start, chrom_end, value]
                print(*row, sep="\t", file=outfile)


def load_posterior_save_bed(coord, resolution, do_reverse,
                            outfilename, num_labels, infilename,
                            num_sublabels=1, output_label="seg"):
    with open(infilename) as infile:
        read_posterior_save_bed(coord, resolution, do_reverse, outfilename,
                                int(num_labels), infile, num_sublabels,
                                output_label)


def parse_viterbi_save_bed(coord, resolution, do_reverse,
                           viterbi_lines, bed_filename, num_labels,
                           output_label):
    data = parse_viterbi(viterbi_lines, do_reverse, output_label)

    start_pos, labels = find_segment_starts(data, output_label)

    save_bed(bed_filename, start_pos, labels, coord, resolution,
             int(num_labels))


def load_viterbi_save_bed(coord, resolution, do_reverse, outfilename,
                          num_labels, infilename):
    with open(infilename) as infile:
        lines = infile.readlines()

    return parse_viterbi_save_bed(coord, resolution, do_reverse,
                                  lines, outfilename, num_labels)


def print_to_fd(fd, line):
    with fdopen(fd, "w") as outfile:
        print(line, file=outfile)


def run_posterior_save_bed(coord, resolution, do_reverse, outfilename,
                           num_labels, num_sublabels, output_label,
                           genomedata_names, distribution, track_indexes_text,
                           virtual_evidence, virtual_evidence_coords,
                           virtual_evidence_priors, *args):
    # XXX: this whole function is duplicative of run_viterbi_save_bed
    # and needs to be reduced convert from tuple
    args = list(args)

    # a 2,000,000-frame output file is only 84 MiB so it is okay to
    # read the whole thing into memory
    (chrom, start, end) = coord
    # Create and save the window
    genomedata_names = genomedata_names.split(",")
    track_indexes = make_track_indexes(track_indexes_text)

    continuous_cells = make_continuous_cells(track_indexes, genomedata_names,
                                             chrom, start, end)

    num_labels = literal_eval(num_labels)
    virtual_evidence_cells = prepare_virtual_evidence(virtual_evidence,
                                                      start, end, num_labels,
                                                      virtual_evidence_coords,
                                                      virtual_evidence_priors)

    temp_filenames = prepare_gmtk_observations(args, chrom, start, end,
                                               continuous_cells,
                                               resolution, distribution,
                                               None, virtual_evidence_cells,
                                               num_labels)
    # remove from memory
    del continuous_cells
    gc.collect()

    with files_to_remove(temp_filenames):
        output = POSTERIOR_PROG.getoutput(*args)

    lines = output.splitlines()
    return read_posterior_save_bed(coord, resolution, do_reverse, outfilename,
                                   int(num_labels), lines,
                                   num_sublabels, output_label)


def run_viterbi_save_bed(coord, resolution, do_reverse, outfilename,
                         num_labels, num_sublabels, output_label,
                         genomedata_names, distribution, track_indexes_text,
                         virtual_evidence, virtual_evidence_coords,
                         virtual_evidence_priors, *args):
    # convert from tuple
    args = list(args)
    # a 2,000,000-frame output file is only 84 MiB so it is okay to
    # read the whole thing into memory

    (chrom, start, end) = coord

    # Create and save the window
    genomedata_names = genomedata_names.split(",")
    track_indexes = make_track_indexes(track_indexes_text)

    continuous_cells = make_continuous_cells(track_indexes, genomedata_names,
                                             chrom, start, end)

    num_labels = literal_eval(num_labels)
    virtual_evidence_cells = prepare_virtual_evidence(virtual_evidence,
                                                      start, end, num_labels,
                                                      virtual_evidence_coords,
                                                      virtual_evidence_priors)

    temp_filenames = prepare_gmtk_observations(args, chrom, start, end,
                                               continuous_cells,
                                               resolution, distribution,
                                               None, virtual_evidence_cells,
                                               num_labels)
    # remove from memory
    del continuous_cells
    gc.collect()

    with files_to_remove(temp_filenames):
        output = VITERBI_PROG.getoutput(*args)

    lines = [line.decode(SEGWAY_ENCODING) for line in output.splitlines()]

    return parse_viterbi_save_bed(coord, resolution, do_reverse,
                                  lines, outfilename, num_labels, output_label)


def run_train(coord, resolution, do_reverse, outfilename,
              genomedata_names, distribution,
              track_indexes,
              is_semisupervised, supervision_coords, supervision_labels,
              virtual_evidence, virtual_evidence_coords,
              virtual_evidence_priors, num_labels, *args):

    # Create and save the train window
    genomedata_names = genomedata_names.split(",")
    track_indexes = make_track_indexes(track_indexes)

    (chrom, start, end) = coord
    gmtk_args = list(args)

    continuous_cells = make_continuous_cells(track_indexes, genomedata_names,
                                             chrom, start, end)

    # XXX: Currently disabled until dinucleotide is enabled
    # Only set these when dinucleotide is set
    # Get the first genome from this world to use for generating
    # sequence cells
    # with Genome(genomedata_names[0]) as genome:
    #     chromosome = genome[chrom]
    #     seq_cells = self.make_seq_cells(chromosome, start, end)

    # If this training regions is supervised
    if is_semisupervised == "True":
        # Create the supervision cells for this region
        # Convert supervision parameters back from text
        supervision_coords = literal_eval(supervision_coords)
        supervision_labels = literal_eval(supervision_labels)

        supervision_cells = make_supervision_cells(supervision_coords,
                                                   supervision_labels,
                                                   start, end)
    else:
        # Otherwise ignore supervision
        supervision_cells = None

    num_labels = literal_eval(num_labels)
    virtual_evidence_cells = prepare_virtual_evidence(virtual_evidence,
                                                      start, end, num_labels,
                                                      virtual_evidence_coords,
                                                      virtual_evidence_priors)

    temp_filenames = prepare_gmtk_observations(gmtk_args, chrom, start,
                                               end, continuous_cells,
                                               resolution, distribution,
                                               supervision_cells,
                                               virtual_evidence_cells,
                                               num_labels)
    del continuous_cells
    gc.collect()

    with files_to_remove(temp_filenames):
        TRAIN_PROG.run(*gmtk_args)


def run_bundle_train(coord, resolution, do_reverse, outfilename, *args):
    args = list(args)

    # Create placeholder observation lists
    placeholder_float_list, placeholder_int_list, \
        placeholder_virtual_evidence_list = \
        save_temp_observation_filelists(PLACEHOLDER_OBSERVATION_FILENAME,
                                        PLACEHOLDER_OBSERVATION_FILENAME,
                                        PLACEHOLDER_OBSERVATION_FILENAME)
    # Modify the given gmtk arguments to use the temporary placeholder
    # observation lists
    replace_subsequent_value(args, EXT_OPTIONS[EXT_FLOAT],
                             placeholder_float_list)
    replace_subsequent_value(args, EXT_OPTIONS[EXT_INT], placeholder_int_list)

    # cppCommandOptions is stored as a string with format CPP_DIRECTIVE_FMT
    cpp_command_options_index = args.index("-cppCommandOptions") + 1
    cpp_command_str = args[cpp_command_options_index]

    # if the placeholder is present, it is replaced. otherwise, the cpp options
    # are unchanged
    args[cpp_command_options_index] = cpp_command_str.replace(
                                    VIRTUAL_EVIDENCE_LIST_FILENAME_PLACEHOLDER,
                                    placeholder_virtual_evidence_list, 1)
    # Run EM bundling
    with files_to_remove([placeholder_float_list, placeholder_int_list]):
        TRAIN_PROG.getoutput(*args)


def save_gmtk_observation_files(coord, resolution, do_reverse, outfile_name,
                                genomedata_names, float_filename,
                                int_filename, distribution,
                                track_indexes, *args):

    (chrom, start, end) = coord

    genomedata_names = genomedata_names.split(",")
    track_indexes = map(int, track_indexes.split(","))

    continuous_cells = make_continuous_cells(track_indexes, genomedata_names,
                                             chrom, start, end)

    _save_window(float_filename, int_filename, continuous_cells,
                 resolution, distribution)


def run_validate(coord, resolution, do_reverse, outfilename, *args):
    if do_reverse:
        raise NotImplementedError("Running Segway with both validation and "
                                  "reverse world options simultaneously is "
                                  "currently not supported")

    validation_output = VALIDATE_PROG.getoutput(*args).decode(SEGWAY_ENCODING)
    with open(outfilename, "w") as outfile:
        outfile.write(validation_output)


TASKS = {("run", "viterbi"): run_viterbi_save_bed,
         ("load", "viterbi"): load_viterbi_save_bed,
         ("run", "posterior"): run_posterior_save_bed,
         ("load", "posterior"): load_posterior_save_bed,
         ("run", "train"): run_train,
         ("run", "bundle-train"): run_bundle_train,
         ("run", "validate"): run_validate,
         ("save", "gmtk-observation-files"): save_gmtk_observation_files,
         }


def task(verb, kind, outfilename, chrom, start, end, resolution,
         reverse, *args):
    start = int(start)
    end = int(end)
    resolution = int(resolution)
    reverse = int(reverse)

    TASKS[verb, kind]((chrom, start, end), resolution, reverse,
                      outfilename, *args)


def main(args=sys.argv[1:]):
    if len(args) < 7:
        print(USAGE, file=sys.stderr)
        sys.exit(2)

    # Try running the task
    try:
        return task(*args)
    # If the there is an explicit out of memory exception
    except MemoryError:
        # return EX_TEMPFAIL error code
        return EX_TEMPFAIL
    # If the viterbi prog returns a non zero exit status
    except optbuild.ReturncodeError as return_code_exception:
        # return the non zero exit status
        return return_code_exception.returncode


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
from __future__ import absolute_import, division

__version__ = "$Revision$"

# Copyright 2009, 2011, 2012 Michael M. Hoffman <michael.hoffman@utoronto.ca>

from math import ceil
import sys

from .._util import MB
from .common import (CLEAN_SAFE_TIME, _JobTemplateFactory,
                     make_native_spec, NULL_FILENAME)

# pbs_drmaa requires that keep_completed be set on the server or queue
# level in order to keep job information and allow its inspection
# (such as for exit status)

class JobTemplateFactory(_JobTemplateFactory):
    # override outputPath and errorPath
    set_template_output_error = False

    def __init__(self, *args, **kwargs):
        _JobTemplateFactory.__init__(self, *args, **kwargs)

        # XXX: Jay says that these must be absolute paths. I think you
        # should be able to prepend drmaa.const.PLACEHOLDER_WD to the
        # path to get it to work, but I can't test this
        self.template.outputPath = NULL_FILENAME
        self.template.errorPath = self.error_filename.abspath()

    def make_res_req(self, mem_usage, tmp_usage):
        # XXX: is there a way to specify tmp_usage in PBS?
        return [make_single_res_req("mem", mem_usage),
                make_single_res_req("vmem", self.mem_limit)]

    def make_native_spec(self):
        # qsub -l: resource requirement
        res_spec = make_native_spec(l=self.res_req)

        res = " ".join([self.native_spec, res_spec])

        return res

def make_single_res_req(name, mem):
    # round up to the next mebibyte
    # http://www.clusterresources.com/torquedocs21/2.1jobsubmission.shtml#size
    return "%s=%dmb" % (name, ceil(mem / MB))


def get_job_max_query_lifetime():
    """
    Get the maximum time in seconds a job's status can be queried (by DRMAA)
    """
    return CLEAN_SAFE_TIME


def main(args=sys.argv[1:]):
    pass


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
from __future__ import absolute_import, division

__version__ = "$Revision$"

# Copyright 2009, 2011, 2012, 2014 Michael M. Hoffman <michael.hoffman@utoronto.ca>

from os import environ
import sys
from tempfile import gettempdir

from path import path

from .._configparser import OneSectionRawConfigParser
from .._util import ceildiv, data_filename
from .common import (BASH_CMD, CLEAN_SAFE_TIME, _JobTemplateFactory,
                     make_native_spec, NULL_FILENAME)

# use SI (not binary) prefixes. I can't find any documentation for
# this but Tim Cutts seems to assume this is how it works
KB = 10**3
MB = 10**6
GB = 10**9
TB = 10**12
PB = 10**15
EB = 10**18

SIZE_UNITS = dict((unit, globals()[unit])
                  for unit in ["KB", "MB", "GB", "TB", "PB", "EB"])

LSF_CONF_BASENAME = "lsf.conf"
LSF_CONF_FILEPATH = path(environ["LSF_ENVDIR"]) / LSF_CONF_BASENAME

# XXX: this can be more robustly handled with the LSLIB function
# ls_readconfenv() and ctypes instead
LSF_CONF = OneSectionRawConfigParser(dict(LSF_UNIT_FOR_LIMITS="KB"))
LSF_CONF.read(LSF_CONF_FILEPATH)

UNIT_FOR_LIMITS = LSF_CONF.get("LSF_UNIT_FOR_LIMITS")
DIVISOR_FOR_LIMITS = SIZE_UNITS[UNIT_FOR_LIMITS]

CORE_FILE_SIZE_LIMIT = 0
HARD_RESOURCE_MULTIPLIER = 2

# this reduces failures due to systems that are simply borked for
# arbitrary reasons
PREEXEC_CMDLINE = "/bin/true"

RES_CLEAN = "segway-clean.sh"
TEMP_DIRNAME = gettempdir()

class JobTemplateFactory(_JobTemplateFactory):
    # eliminate default overwrite behavior for DRMAA/LSF, go to append
    # which is default for DRMAA/SGE
    # we use a native spec -o and -e instead of the DRMAA interface
    set_template_output_error = False

    def make_res_req(self, mem_usage, tmp_usage):
        mem_usage_spec = ceildiv(mem_usage, DIVISOR_FOR_LIMITS)
        tmp_usage_spec = ceildiv(tmp_usage, DIVISOR_FOR_LIMITS)

        # returns a quoted string
        return ('"select[mem>%s && tmp>%s] rusage[mem=%s, tmp=%s]"'
                % (mem_usage_spec, tmp_usage_spec,
                   mem_usage_spec, tmp_usage_spec))

    def make_postexec_cmdline(self):
        assert TEMP_DIRNAME.startswith("/")

        res = [BASH_CMD, data_filename(RES_CLEAN)]
        for arg in self.args:
            if arg.startswith(TEMP_DIRNAME):
                res.append(arg)

        # returns a quoted string
        return '"%s"' % " ".join(res)

    def make_native_spec(self):
        mem_limit_spec = ceildiv(self.mem_limit, DIVISOR_FOR_LIMITS)

        # these all use the unit defined by LSF_UNIT_FOR_LIMITS

        # see Administering Platform LSF: Controlling Job Execution:
        # Runtime Resource Usage Limits: Scaling the units for
        # resource usage limits

        # bsub -R: resource requirement
        # bsub -M: per-process memory limit
        # bsub -v: hard virtual memory limit for all processes, cumulatively
        # bsub -C: core file size limit
        res_spec = make_native_spec(R=self.res_req, M=mem_limit_spec,
                                    v=mem_limit_spec,
                                    o=NULL_FILENAME,
                                    e=self.error_filename,
                                    E=PREEXEC_CMDLINE,
                                    Ep=self.make_postexec_cmdline())

        # XXX: -C does not work with DRMAA for LSF 1.03
        # wait for upstream fix:
        # https://sourceforge.net/tracker/?func=detail&aid=2882034&group_id=206321&atid=997191
        #                            C=CORE_FILE_SIZE_LIMIT)
        # XXX: need to ping this

        return " ".join([self.native_spec, res_spec])


def get_job_max_query_lifetime():
    """
    Get the maximum time in seconds a job's status can be queried (by DRMAA)
    """
    return CLEAN_SAFE_TIME


def main(args=sys.argv[1:]):
    pass


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
from __future__ import absolute_import, division

__version__ = "$Revision$"

# common stuff: needs to be in a different file from __init__ because
# it is imported by lsf.py and sge.py
# Copyright 2009, 2011, 2012, 2013 Michael M. Hoffman <michael.hoffman@utoronto.ca>

import sys
from tempfile import gettempdir

from optbuild import Mixin_NoConvertUnderscore, OptionBuilder_ShortOptWithSpace

from .._util import data_filename, KB, MB

NATIVE_SPEC_PROG = (Mixin_NoConvertUnderscore
                    + OptionBuilder_ShortOptWithSpace)() # do not run

# CLEAN_PERIOD in lsb.params, after which jobs are removed from
# mbatchd's memory default is 3600, multiplying by 0.5 for a margin of
# error
# XXX: check lsb.params for real value of CLEAN_PERIOD
CLEAN_SAFE_TIME = int(3600 * 0.9)

# guard space to prevent going over mem_requested allocation
MEM_GUARD = 10*MB

BASH_CMD = "bash"
RES_WRAPPER = "segway-wrapper.sh"

MSG_EDGE = """
Ran through entire memory usage progression without success.
For details, check error messages in %s.
See the Troubleshooting section of the Segway documentation."""

NULL_FILENAME = "/dev/null"

class JobError(RuntimeError):
    pass

class _JobTemplateFactory(object):
    # this might be overridden by a subclass
    set_template_output_error = True

    def __init__(self, template, tmp_usage, mem_usage_progression,
                 output_filename, error_filename):
        self.args = template.args
        self.native_spec = template.nativeSpecification
        self.mem_usage_progression = mem_usage_progression

        # set here so that segway.cluster.lsf can use it
        self.output_filename = output_filename
        self.error_filename = error_filename

        # only do this if the subclass doesn't do it for you
        if self.set_template_output_error:
            # format:
            # hostname:filename
            # no hostname = execution host
            # XXX: should add drmaa.const.PLACEHOLDER_WD to
            # beginning of relative paths--those that don't start with "/"
            template.outputPath = ":" + NULL_FILENAME
            template.errorPath = ":" + error_filename

        self.template = template
        self.tmp_usage = tmp_usage

    def __call__(self, trial_index):
        """
        returns a job template with the attributes set for the next
        memory progression step
        """
        res = self.template

        try:
            mem_usage = self.mem_usage_progression[trial_index]
        except IndexError:
            raise RuntimeError(MSG_EDGE % self.error_filename)

        self.mem_limit = int(mem_usage)
        self.res_req = self.make_res_req(mem_usage, self.tmp_usage)

        res.args = self.make_args()
        res.nativeSpecification = self.make_native_spec()

        return res

    def make_res_req(self, mem_usage, tmp_usage):
        """
        pure virtual function to be replaced in subclass

        should set self.res_req; return value undefined

        mem_usage: expected virtual memory use in bytes
        tmp_usage: expected temporary space use in bytes
        """
        raise NotImplementedError

    def make_args(self):
        """
        wrap args with segway-wrapper.sh
        """
        # ulimit args are in kibibytes
        mem_limit_kb = str(calc_mem_limit(self.mem_limit) // KB)
        wrapper_cmdline = [BASH_CMD, data_filename(RES_WRAPPER), mem_limit_kb,
                           gettempdir(), self.output_filename]

        return wrapper_cmdline + self.args

    def make_native_spec(self):
        """
        pure virtual function to be replaced in subclass

        should use self.res_req and return a string to go into
        template.nativeSpecification
        """
        raise NotImplementedError

    def delete_job_template(self, session):
        session.deleteJobTemplate(self.template)

def make_native_spec(*args, **kwargs):
    return " ".join(NATIVE_SPEC_PROG.build_args(args=args, options=kwargs))

def calc_mem_limit(mem_usage):
    """
    amount of memory allowed to be used (less than that requested)
    """
    return int(mem_usage - MEM_GUARD)


def main(args=sys.argv[1:]):
    pass

if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
from __future__ import absolute_import, division, print_function

"""observations.py: prepare and save GMTK observations
"""

__version__ = "$Revision$"

# Copyright 2012, 2013 Michael M. Hoffman <michael.hoffman@utoronto.ca>
from collections import deque
from contextlib import closing
from functools import partial
from itertools import repeat
from operator import itemgetter
from os import extsep
import sys
from tempfile import gettempdir
from warnings import warn

from genomedata import Genome
from numpy import (add, any, append, arange, arcsinh, argmax, array,
                   bincount, clip, column_stack, copy, empty, full, invert,
                   isnan, maximum, mean, sum, where, zeros)
from numpy import sum as numpy_sum
from path import Path
from six import viewitems
from six.moves import map, range, StringIO, zip
from tabdelim import ListWriter

from ._util import (ceildiv, copy_attrs, DISTRIBUTION_ASINH_NORMAL,
                    DTYPE_OBS_INT, EXT_FLOAT, EXT_INT, extjoin,
                    get_chrom_coords, make_prefix_fmt, SegwayWarning,
                    SUPERVISION_LABEL_OFFSET, USE_MFSDG, Window)

# number of frames in a segment must be at least number of frames in model
MIN_FRAMES = 2

MAX_WINDOWS = 9999

FLOAT_TAB_FIELDNAMES = ["filename", "window_index", "chrom", "start", "end"]

ORD_A = ord("A")
ORD_C = ord("C")
ORD_G = ord("G")
ORD_T = ord("T")
ORD_a = ord("a")
ORD_c = ord("c")
ORD_g = ord("g")
ORD_t = ord("t")

DIM_TRACK = 1  # Dimension in numpy array for track data

PRIOR_AXIS = 0
POSITION_AXIS = 1

class VirtualEvidenceWarning(SegwayWarning):
    """
    User-supplied priors sum to greater than one.
    """


class NoData(object):
    """
    sentinel for not adding an extra field to coords, so that one can
    still use None
    """


def convert_windows(attrs, name):
    supercontig_start = attrs.start
    edges_array = getattr(attrs, name) + supercontig_start

    return edges_array.tolist()


def merge_windows(windows):
    """Takes a sorted list of start/end tuples and returns a list of tuples of
    merged together if any overlap"""

    res = []

    # Get first region
    merge_start, merge_end = windows[0]

    # For all remaining regions
    for window_start, window_end in windows[1:]:
        # If the next region does not overlap the current one under
        # consideration
        if window_start - merge_end > 0:
            # Add our current merged region to the merged list
            res.append((merge_start, merge_end))
            # Set the current region under consideration to the current window
            merge_start, merge_end = window_start, window_end
        # Otherwise the regions overlap
        else:
            if window_end > merge_end:
                merge_end = window_end

    # Append the last region under consideration to the merge list
    res.append((merge_start, merge_end))
    # Return the list of merged windows
    return res


def update_starts(starts, ends, new_starts, new_ends):
    # reversed because extend left extends deque in reversed order
    starts.extendleft(reversed(new_starts))
    ends.extendleft(reversed(new_ends))


def intersect_regions(start, end, coords, data=repeat(NoData)):
    """
    find items in coords that overlap (start, end)

    NOTE: multiple overlapping regions in coords will result in data
    being considered more than once
    """
    # diagram of how this works:
    #        --------------------- (include)
    #
    # various (start, end) cases:
    # A   --
    # B                              --
    # C   --------------------------
    # D   ------
    # E                       -------------
    # F             ---------
    res = []

    for (include_start, include_end), datum in zip(coords, data):
        if start > include_end or end <= include_start:
            # cases A, B
            continue
        elif start <= include_start:
            if end < include_end:
                # case D
                include_end = end

            # case C otherwise
        elif start > include_start:
            include_start = start

            if end < include_end:
                # case F
                include_end = end

            # case E otherwise
        else:
            assert False  # can't happen

        item = [include_start, include_end]

        if datum is not NoData:
            item.append(datum)

        res.append(item)

    return res


def subtract_regions(start, end, exclude_coords):
    """
    takes a start and end and removes everything in exclude_coords
    """
    # diagram of how this works:
    #        --------------------- (include)
    #
    # various exclude cases:
    # A   --
    # B                              --
    # C   --------------------------
    # D   ------
    # E                       -------------
    # F             ---------

    # does nothing if exclude_coords is empty

    if exclude_coords is None or len(exclude_coords) == 0:
        return [[start, end]]

    include_coords = [[start, end]]

    for exclude_start, exclude_end in exclude_coords:
        new_include_coords = []

        for include_coord in include_coords:
            start, end = include_coord

            if exclude_start > end or exclude_end <= start:
                # cases A, B
                new_include_coords.append([start, end])
            elif exclude_start <= start:
                if exclude_end >= end:
                    # case C
                    pass
                else:
                    # case D
                    new_include_coords.append([exclude_end, end])
            elif exclude_start > start:
                if exclude_end >= end:
                    # case E
                    new_include_coords.append([start, exclude_start])
                else:
                    # case F
                    new_include_coords.append([start, exclude_start])
                    new_include_coords.append([exclude_end, end])

            else:
                assert False  # can't happen

        include_coords = new_include_coords

    return include_coords


def calc_downsampled_shape(inarray, resolution):
    full_num_rows = inarray.shape[0]
    downsampled_num_rows = ceildiv(full_num_rows, resolution)
    return [downsampled_num_rows] + list(inarray.shape[1:])


def downsample_add(inarray, resolution):
    """
    Downsample a matrix by rows to a desired resolution, by adding up
    data points

    [1 2 4 8 16 32] downsampled to resolution 2 is:
    [1+2 4+8 16+32]

    originally: downsample presence data into num_datapoints
    now used for all downsampling

    inarray: array to be downsampled
    resolution: desired resolution

    """
    if resolution == 1:
        return inarray

    full_num_rows = inarray.shape[0]
    res = zeros(calc_downsampled_shape(inarray, resolution), inarray.dtype)

    # if there's no remainder, then only use loop 0
    remainder = full_num_rows % resolution
    if remainder == 0:
        remainder = resolution

    # loop 0: every index up to remainder
    for index in range(remainder):
        res += inarray[index::resolution]

    # loop 1: remainder
    for index in range(remainder, resolution):
        # don't include the last element of res
        res[:-1] += inarray[index::resolution]

    return res


def get_downsampled_supervision_data_and_presence(input_array, resolution):
    """
    Downsample a 1-dimensional numpy array to a desired resolution
    by taking the mode of each resolution-sized frame.
    For example, [1 2 2 3 4 4] downsampled to resolution 3 is: [2 4]

    This function returns a tuple (where each element is an array) of
    the form: [downsampled input array], [presence of downsampled
    input array] where 'presence' is the count of each mode in its
    'window'.

    Only downsamples 1D numpy arrays.

    There is the possibility that there will be a 'remainder'
    subarray. For that case, we choose to take its mode and append
    it to the end of our array of modes.

    As a rule, we will never choose 0 (no label) unless all elements
    of the subarray are 0.

    For example, for [1, 0, 0,...,0] our 'mode' will be 1.
    But [0,0,0,0...,0] has a mode of 0.

    """
    if resolution == 1:
        # at resolution==1, the presence for all nonzero values is 1
        # and the presence for 0 is 0 (no supervision data)
        presence_array = clip(input_array, 0, 1)
        return input_array, presence_array

    # split input_array into subarrays of length resolution.
    # e.g. [1, 1, 3, 4, 5, 6] to [[1, 1, 3], [4, 5, 6] for resolution == 3
    resolution_partitioned_input_array = (
            input_array[index:index+resolution]
            for index in range(0, len(input_array), resolution)
            )

    downsampled_input_array = zeros(calc_downsampled_shape(input_array,
                                    resolution), input_array.dtype)
    presence_downsampled_input_array = zeros(calc_downsampled_shape(input_array,
                                             resolution), input_array.dtype)
    # For each input partition find its mode
    for index, input_partition in enumerate(resolution_partitioned_input_array):
        # Get the number of times each index occurs in the input partition.
        # bincount(a) returns an array where the elements are the number
        # of times each index occurs in a.
        resolution_sized_subarray_bincount = bincount(input_partition)

        # by setting the 0th value of bincount to 0, we take care
        # of two possible special cases:
        # 1) number of 0s is the same as the count of the nonzero mode:
        #   >we want to have the nonzero mode, not 0.
        #   setting count of 0 to 0 will force argmax to default to
        #   the nonzero mode.
        #
        # 2) the case of all 0s (no nonzero mode):
        #   >we want the mode to be 0.
        #   setting count of 0 to 0 will return an argmax of 0 since
        #   there are no other numbers to choose.
        resolution_sized_subarray_bincount[0] = 0

        # in the case of ties for the max count, argmax takes the
        # smallest numbered index.
        mode = argmax(resolution_sized_subarray_bincount)
        downsampled_input_array[index] = mode

        # Get the count of the mode in our input partition
        presence_downsampled_input_array[index] = \
            resolution_sized_subarray_bincount[mode]

    return downsampled_input_array, presence_downsampled_input_array


def downsample_presence_array(presence_array, resolution):
    """
    Downsamples a 1D presence array given the specified resolution.
    Sums the presence across bins given the resolution size, so
    [0, 1, 1, 1, 0, 0] for resolution == 3 becomes
    [2, 1]
    """
    return [numpy_sum(presence_array[index:index+resolution])
            for index in range(0, len(presence_array), resolution)]


def downsample_prior_array(raw_prior_array, resolution, uniform_priors):
    """
    Downsamples a 2D array of priors. Takes the average for all labels defined,
    uses a uniform prior if none are.
    """
    # split input_array into subarrays of length resolution.
    # e.g. [1,1,3,4,5,6] to [[1,1,3],[4,5,6]] for resolution == 3
    resolution_partitioned_prior_gen = (
            raw_prior_array[index:index+resolution]
            for index in range(0, len(raw_prior_array), resolution))

    # Empty array to be filled with mean values for priors for each bin
    res = empty(calc_downsampled_shape(raw_prior_array,
                                       resolution), raw_prior_array.dtype)

    # For each input partition, calculate the mean prior for each label
    # if no priors given, use a uniform prior
    for index, input_partition in enumerate(resolution_partitioned_prior_gen):
        # if no priors are defined in this partition,
        # (meaning input_partition will be entirely composed of 0s)
        # set the mean vector to be uniform
        if not input_partition.any():
            res[index] = uniform_priors
        else:
            # Remove any rows which don't contain priors
            defined_priors = input_partition[input_partition.any(POSITION_AXIS)]

            # take the mean of the given priors for each label over the
            # position axis
            mean_prior_vector = mean(defined_priors, axis=PRIOR_AXIS)

            res[index] = mean_prior_vector

    return res

def get_downsampled_virtual_evidence_data_and_presence(raw_prior_array,
                                                       resolution, num_labels):
    """
    Downsample a 2-dimensional array of label probabilities to a
    desired resolution by taking the average for all labels
    over the positions with priors.

    If no priors are specified then the average prior is uniform.

    This function returns a tuple (where each element is an array) of
    the form: [downsampled input array], [presence of downsampled
    input array] where 'presence' is the count of evidence-specified positions
    in each subpartition.

    Downsampled input array is an array of arrays where each element
    is the vector of priors for the given downsampled position.

    There is the possibility that there will be a 'remainder'
    subarray. For that case, we choose to take its average (same as above)
    and append it to the end of our array of downsampled priors.

    For example, the 3-position frame

    [[0.5, 0.2, 0.3],
     [0.3, 0.4, 0.3],
     [0.0, 0.0, 0.0]]

    downsampled to resolution 3 is [0.4, 0.3, 0.3]
    since for label 0 we take mean(0.5, 0.3) = 0.4
          for label 1 we take mean(0.2, 0.4) = 0.3
          for label 2 we take mean(0.3, 0.3) = 0.3
    """

    uniform_priors = array([1.0/num_labels] * num_labels)

    # take the presence to be 1 at every position the user has defined
    # any priors and 0 otherwise
    presence_array = raw_prior_array.any(axis=POSITION_AXIS).astype(DTYPE_OBS_INT)

    if resolution == 1:
        # our "downsampled" prior array at resolution 1 is just the
        # vector of priors defined by the user at every position
        # with uniform priors filled in at all other positions
        prior_array = zeros((len(raw_prior_array), num_labels))
        for prior_list_index, prior_vector in enumerate(raw_prior_array):
            if prior_vector.sum() == 0:
                prior_array[prior_list_index] = uniform_priors
            else:
                prior_array[prior_list_index] = prior_vector

        return prior_array, presence_array

    downsampled_prior_array = downsample_prior_array(raw_prior_array,
                                                     resolution,
                                                     uniform_priors)
    downsampled_presence_array = downsample_presence_array(presence_array,
                                                           resolution)


    return downsampled_prior_array, downsampled_presence_array


def make_dinucleotide_int_data(seq):
    """
    makes an array with two columns, one with 0..15=AA..TT and the other
    as a presence variable. Set column one to 0 when not present
    """
    nucleotide_int_data = (((seq == ORD_A) + (seq == ORD_a))
                           + ((seq == ORD_C) + (seq == ORD_c)) * 2
                           + ((seq == ORD_G) + (seq == ORD_g)) * 3
                           + ((seq == ORD_T) + (seq == ORD_t)) * 4) - 1
    nucleotide_missing = nucleotide_int_data == -1

    # rewrite all Ns as A now that you have the missingness mask
    nucleotide_int_data[nucleotide_int_data == -1] = 0
    col_shape = (len(nucleotide_int_data) - 1,)

    # first column: dinucleotide: AA..TT=0..15
    # combine, and add extra AA stub at end, which will be set missing
    # 0 AA AC AG AT
    # 4 CA CC CG CT
    # 8 GA GC GG GT
    # 12 TA TC TG TT
    dinucleotide_int_data = empty(col_shape, DTYPE_OBS_INT)
    add(nucleotide_int_data[:-1] * 4, nucleotide_int_data[1:],
        dinucleotide_int_data)

    # second column: presence_dinucleotide: some missing=0; all present = 1
    # there are so few N boundaries that it is okay to
    # disregard the whole dinucleotide when half is N
    dinucleotide_missing = (nucleotide_missing[:-1] + nucleotide_missing[1:])

    dinucleotide_presence = empty(dinucleotide_missing.shape, DTYPE_OBS_INT)
    invert(dinucleotide_missing, dinucleotide_presence)

    # XXXopt: set these up properly in the first place instead of
    # column_stacking at the end
    return column_stack([dinucleotide_int_data, dinucleotide_presence])


def make_supervision_cells(supervision_coords, supervision_labels, start, end):
    """
    supervision_coords: list of tuples (start, end)
    supervision_labels: list of ints (label as number)
    start: int
    end: int

    returns a 1-dimensional numpy.ndarray for the region specified by
    superivion_coords where each cell is the transformed label specified by
    supervision_labels for that region

    the transformation results in the cell being 0 for no supervision
    or SUPERVISION_LABEL_OFFSET (1)+the supervision label for supervision
    """

    res = zeros(end - start, dtype=DTYPE_OBS_INT)

    # Get supervision regions that overlap with the start and end coords
    supercontig_coords_labels = \
        intersect_regions(start, end, supervision_coords,
                          supervision_labels)

    for label_start, label_end, label_index in supercontig_coords_labels:
        # adjust so that zero means no label
        label_adjusted = label_index + SUPERVISION_LABEL_OFFSET
        res[(label_start - start):(label_end - start)] = label_adjusted

    return res


def fill_virtual_evidence_cells(prior_input_array, num_labels):
    """
    For genomic positions which have at least one, but not all priors specified,
    this function will apply a uniform prior to all remaining labels.
    Indexes where no prior was ever specified will remain zero for downsampling

    Example:
    INPUT:
    [[0.5, 0.2, None, None, None],
     [None, 0.4, None, None, None],
     [None, None, None, None, None]]

    OUTPUT: 
    [[0.5, 0.2, 0.10, 0.10, 0.10],
     [0.15, 0.4, 0.15, 0.15, 0.15],
     [0.0 ... 0.0]]
    """

    prior_array = zeros((len(prior_input_array), num_labels))
    for index, prior_input in enumerate(prior_input_array):
        # Only priors which were specified in the input file will be set
        # Unset labels will still be none
        num_prior_labels = numpy_sum(prior_input != None)
        if num_prior_labels:
            prior_list_values = list(filter(None, prior_input))

            # Check if priors should be treated as ratios or percentages
            if numpy_sum(prior_list_values) < 1:
                remaining_probability = 1 - numpy_sum(prior_list_values)
            else:
                remaining_probability = 0

            # divide remaining probability uniformly amongst the remaining labels
            prior_input[prior_input == None] = (remaining_probability /
                                                (num_labels-num_prior_labels))

            prior_array[index] = prior_input

    return prior_array

def check_is_close(prior_sums, reference):
    return all((prior_sums - reference) <= 0)

def make_virtual_evidence_cells(coords, priors,
                                start, end, num_labels):
    """
    coords: list of tuples (start, end)
    priors: list of dictionaries of the format {label: prior}
    start: int
    end: int

    returns a 2-dimensional numpy.ndarray for the section of the array bounded
    by start and end where each cell is the prior data for each label of each
    region inside the section.
    """
    res = full((num_labels, (end-start)), None)

    # Get supervision regions that overlap with the start and end coords
    supercontig_coords_labels = \
        intersect_regions(start, end, coords, priors)

    for label_start, label_end, prior_dict in supercontig_coords_labels:
        label = list(prior_dict.keys())[0]
        # copy data to all positions
        if res[label][(label_start - start):(label_end - start)].any():
            raise ValueError("VE label {} overlaps in coordinates {}-{}".format(label,
                             label_start, label_end))
        res[label][(label_start - start):(label_end - start)] = \
            list(prior_dict.values()) * ((label_end - start)-(label_start - start))

    # For coords which had at least one label supplied, fill remaining labels
    # with uniform remaining probability
    # Transpose the priors so that it has positions as rows and labels as columns
    res = fill_virtual_evidence_cells(res.transpose(), num_labels)

    if any(res < 0):
        raise ValueError("Priors may not be negative")

    is_close = check_is_close(res.sum(axis=POSITION_AXIS), 1.0)
    if not is_close:
        close_indexes = where(res.sum(axis=POSITION_AXIS) > 1.0)[0]
        warn("Prior labels sum to greater than one in on genomic indexes {}".format(close_indexes + start),
             VirtualEvidenceWarning)

    return res


def make_continuous_cells(track_indexes, genomedata_names,
                          chromosome_name, start, end):
    """
    returns 2-dimensional numpy.ndarray of continuous observation
    data for specified interval. This data is untransformed

    dim 0: position
    dim 1: track
    """

    continuous_cells = None

    # For every track in each genomedata archive
    zipper = zip(track_indexes, genomedata_names)
    for track_index, genomedata_name in zipper:
        with Genome(genomedata_name) as genome:
            chromosome = genome[chromosome_name]
            # If we haven't started creating the continous cells
            if continuous_cells is None:
                # Copy the first track into our continous cells
                continuous_cells = copy(chromosome[start:end,
                                        [track_index]])
            else:
                # Otherwise append the track to our continuous cells
                continuous_cells = append(continuous_cells,
                                          chromosome[start:end,
                                                     [track_index]],
                                          DIM_TRACK)

    return continuous_cells


def _save_window(float_filename_or_file, int_filename_or_file, 
                 float_data, resolution, distribution, seq_data=None,
                 supervision_data=None, virtual_evidence_data=None,
                 virtual_evidence_filename_or_file=None, num_labels=None):
    # called by task.py as well as observation.py

    # input function in GMTK_ObservationMatrix.cc:
    # ObservationMatrix::readBinSentence

    # Input per frame is a series of float32s, followed by a series of
    # int32s. It is better to optimize both sides here by sticking all
    # the floats in one file, and the ints in another one.
    int_blocks = []
    if float_data is not None:
        if distribution == DISTRIBUTION_ASINH_NORMAL:
            float_data = arcsinh(float_data)

        if (not USE_MFSDG) or resolution > 1:
            mask_missing = isnan(float_data)

            # output -> presence_data -> int_blocks
            # done in two steps so I can specify output type
            presence_data = empty(mask_missing.shape, DTYPE_OBS_INT)
            invert(mask_missing, presence_data)

            num_datapoints = downsample_add(presence_data, resolution)

            # this is the presence observation
            int_blocks.append(num_datapoints)

            # so that there is no divide by zero
            num_datapoints_min_1 = maximum(num_datapoints, 1)

            # make float
            if not USE_MFSDG:
                float_data[mask_missing] = 0.0

            float_data = downsample_add(float_data, resolution)
            float_data /= num_datapoints_min_1

        float_data.tofile(float_filename_or_file)

    if seq_data is not None:
        assert resolution == 1  # not implemented yet
        int_blocks.append(make_dinucleotide_int_data(seq_data))

    if supervision_data is not None:
        supervision_data, presence_supervision_data = \
            get_downsampled_supervision_data_and_presence(supervision_data,
                                                          resolution)

        int_blocks.append(supervision_data)
        int_blocks.append(presence_supervision_data)

    if virtual_evidence_data is not None:
        virtual_evidence_data_array, presence_virtual_evidence_data = \
            get_downsampled_virtual_evidence_data_and_presence(
                virtual_evidence_data,
                resolution,
                num_labels)
        int_blocks.append(presence_virtual_evidence_data)

        # separately save VE priors CPT in a temporary file
        # done using write since ndarray.tofile does not format correctly
        # Produces error: observation file 0 '/tmp/segway.qvHHEnnQV4/ve.gradckoq.list' segment 0: couldn't read 0'th item in frame 0
        for prior in virtual_evidence_data_array:
            virtual_evidence_filename_or_file.write(
                ' '.join(['{}'.format(prob) for prob in prior]) + '\n')

    if int_blocks:
        int_data = column_stack(int_blocks).astype(dtype=DTYPE_OBS_INT)
    else:
        int_data = array([], dtype=DTYPE_OBS_INT)

    int_data.tofile(int_filename_or_file)

def add_starts_ends(new_windows, starts, ends):
    """
    If there are no new windows, return None, None

    If there is more than one new window, modify the
    starts and ends in place and return None, None

    If there is only one new window, return a tuple of
    (start, end) for that window.

    """
    if not new_windows:  # nothing left
        return None, None
    elif len(new_windows) > 1:
        new_starts, new_ends = zip(*new_windows)
        # reversed because extend left extends deque in reversed order
        starts.extendleft(reversed(new_starts))
        ends.extendleft(reversed(new_ends))
        return None, None

    return new_windows[0]

def generate_coords_from_dict(coords_dict):
    # use a deque to allow fast insertion/removal at the
    # beginning and end of the sequence
    for chrom, coords_list in viewitems(coords_dict):
        starts, ends = map(deque, zip(*coords_list))
        yield chrom, starts, ends

class Observations(object):
    copy_attrs = ["include_coords", "exclude_coords", "max_frames",
                  "float_filelistpath", "int_filelistpath",
                  "validation_float_filelistpath",
                  "validation_int_filelistpath",
                  "float_tabfilepath", "validation_obs_dirpath",
                  "obs_dirpath", "uuid", "resolution",
                  "distribution", "train", "identify", "random_state",
                  "supervision_type",
                  "supervision_coords", "supervision_labels",
                  "use_dinucleotide", "world_track_indexes",
                  "world_genomedata_names", "clobber",
                  "num_worlds", "validation_fraction", "validate",
                  "validation_coords", "virtual_evidence"]

    def __init__(self, runner):
        copy_attrs(runner, self, self.copy_attrs)

        self.float_filepaths = []
        self.int_filepaths = []

        self.validation_float_filepaths = []
        self.validation_int_filepaths = []

        self.validation_windows = []

    def generate_coords_all(self, genomes):
        """Generates a tuple of (chromosome, starts, ends) for each chromosome
        across all genomes where the starts and ends are deques"""

        # Get all chromosome names across all genomes
        chromosome_names = set()
        for genome in genomes:
            chromosome_names = chromosome_names.union([chromosome.name for
                                                       chromosome in genome])

        # For each chromsome (in sorted order)
        for chromosome_name in sorted(list(chromosome_names)):
            chromosome_windows = []  # list of (start, end) tuples

            # Get all genome-mapped regions across all genomes
            for genome in genomes:
                chromosome = genome[chromosome_name]

                # For each supercontig
                for supercontig, continuous in chromosome.itercontinuous():
                    # If the the supercontig is not empty
                    if continuous is not None:
                        attrs = supercontig.attrs

                        starts = convert_windows(attrs, "chunk_starts")
                        ends = convert_windows(attrs, "chunk_ends")

                        chromosome_windows.extend(zip(starts, ends))

            # If at least 1 window exists
            if chromosome_windows:
                # Sort regions (by start)
                chromosome_windows.sort(key=itemgetter(0))
                # Merge regions
                merged_chromsome_windows = merge_windows(chromosome_windows)
                # Convert start and end regions to deques
                starts, ends = map(deque, zip(*merged_chromsome_windows))
                # Yield the chromsome name and start/end region deques
                yield chromosome_name, starts, ends

    def generate_coords(self, genomes):
        """
        returns iterable of included coords, either explicitly
        specified, or all

          each item: tuple of (chrom, starts, ends)
          starts and ends are deques
        """
        if self.include_coords:
            return generate_coords_from_dict(self.include_coords)
        else:
            return self.generate_coords_all(genomes)

    def skip_or_split_window(self, start, end):
        """
        skip short windows or skip long windows
        """
        max_frames = self.max_frames
        num_bases_window = end - start
        num_frames = ceildiv(num_bases_window, self.resolution)
        if not MIN_FRAMES <= num_frames:
            text = " skipping short sequence of length %d" % num_frames
            print(text, file=sys.stderr)
            return []

        if num_frames > max_frames:
            # XXX: should check that this is going to always work even for
            # corner cases (what corner cases?), but if it doesn't,
            # another split later on fixes it

            # split_sequences was True, so split them
            num_new_starts = ceildiv(num_frames, max_frames)

            # // means floor division
            offset = (num_frames // num_new_starts)
            new_offsets = arange(num_new_starts) * (offset * self.resolution)
            new_starts = start + new_offsets
            new_ends = append(new_starts[1:], end)

            return list(zip(new_starts, new_ends))

        return [[start, end]]

    def create_validation_windows_from_fraction(self, windows):
        """
        Takes a subset of the training windows, using the specified
        validation fraction, to be used as the validation windows.

        Note: modifies the input parameter 'windows' and uses this
        as the training windows.

        Returns training windows, validation windows
        """
        validation_windows = []
        total_bases = sum(len(window) for window in windows)
        cur_bases = 0
        self.random_state.shuffle(windows)

        # Remove windows from training set and add them to validation set
        # until total number of bases chosen is at least
        # as large as that required by the validation fraction
        while (float(cur_bases) / total_bases) < self.validation_fraction:
            window = windows.pop()
            validation_windows.append(window)
            cur_bases += len(window)

        return windows, validation_windows

    def create_validation_windows_from_coords(self, exclude_coords):
        """
        Creates set of validation windows using the specified
        validation coordinates.

        Returns validation windows
        """
        validation_windows = []
        for chrom, starts, ends in generate_coords_from_dict(self.validation_coords):
            chr_exclude_coords = get_chrom_coords(exclude_coords, chrom)

            while len(starts) > 0:
                start = starts.popleft()
                end = ends.popleft()  # should not ever cause an IndexError

                new_validation_windows = \
                    subtract_regions(start, end, chr_exclude_coords)
                subtracted_start, subtracted_end = \
                    add_starts_ends(new_validation_windows,
                                        starts,
                                        ends)
                if subtracted_start:
                    # skip or split long sequences
                    new_validation_windows = \
                        self.skip_or_split_window(subtracted_start,
                                                  subtracted_end)
                    split_start, split_end = \
                        add_starts_ends(new_validation_windows,
                                            starts,
                                            ends)
                    if split_start:
                        for world in range(self.num_worlds):
                            validation_windows.append(Window(
                                world, chrom, split_start, split_end)
                                )

        return validation_windows

    def locate_windows(self, genomes):
        """
        input: Genome instances, include_coords, exclude_ coords, max_frames
        validation fraction/validation coords (if validation)

        sets: window_coords, validation_coords (if validation)
        """
        exclude_coords = self.exclude_coords

        windows = []

        for chrom, starts, ends in self.generate_coords(genomes):
            chr_exclude_coords = get_chrom_coords(exclude_coords, chrom)

            while True:
                try:
                    start = starts.popleft()
                except IndexError:
                    break

                end = ends.popleft()  # should not ever cause an IndexError

                new_windows = subtract_regions(start, end, chr_exclude_coords)
                start, end = add_starts_ends(new_windows, starts, ends)
                if start is None:
                    continue

                # skip or split long sequences
                new_windows = self.skip_or_split_window(start, end)
                start, end = add_starts_ends(new_windows, starts, ends)
                if start is None:
                    continue

                for world in range(self.num_worlds):
                    windows.append(Window(world, chrom, start, end))

        if self.validation_fraction:
            windows, self.validation_windows = \
                self.create_validation_windows_from_fraction(windows)
        elif self.validation_coords:
            self.validation_windows = \
                self.create_validation_windows_from_coords(exclude_coords)

        if not windows:
            raise ValueError("Set of training windows is empty")

        # remaining windows can be passed into training now
        self.windows = windows

    def save_window(self, float_filename, int_filename, float_data,
                    seq_data=None, supervision_data=None):
        return _save_window(float_filename, int_filename, float_data,
                            self.resolution, self.distribution, seq_data,
                            supervision_data)

    @staticmethod
    def make_filepath(dirpath, prefix, suffix):
        return dirpath / (prefix + suffix)

    def make_validation_filepaths(self, chrom, window_index):
        prefix_feature_tmpl = extjoin(chrom, make_prefix_fmt(MAX_WINDOWS))
        prefix = prefix_feature_tmpl % window_index

        dirpath = self.validation_obs_dirpath

        make_filepath_custom = partial(self.make_filepath, dirpath, prefix)

        return (make_filepath_custom(EXT_FLOAT), make_filepath_custom(EXT_INT))

    def make_filepaths(self, chrom, window_index, temp=False):
        prefix_feature_tmpl = extjoin(chrom, make_prefix_fmt(MAX_WINDOWS))
        prefix = prefix_feature_tmpl % window_index

        if temp:
            prefix = "".join([prefix, self.uuid, extsep])
            dirpath = Path(gettempdir())
        else:
            dirpath = self.obs_dirpath

        make_filepath_custom = partial(self.make_filepath, dirpath, prefix)

        return (make_filepath_custom(EXT_FLOAT), make_filepath_custom(EXT_INT))

    def print_filepaths(self, float_filelist, int_filelist, *args, **kwargs):
        float_filepath, int_filepath = self.make_filepaths(*args, **kwargs)
        print(float_filepath, file=float_filelist)
        print(int_filepath, file=int_filelist)

        self.float_filepaths.append(float_filepath)
        self.int_filepaths.append(int_filepath)

        return float_filepath, int_filepath

    def create_filepaths(self, temp=False):
        """ Creates a list of observations full filepaths for each window.
        temp is a flag to determine whether or not the filepaths will be
        given a temporary directory """

        for window_index, window in enumerate(self.windows):
            float_filepath, int_filepath = self.make_filepaths(
                                            window.chrom, window_index,
                                            temp)
            self.float_filepaths.append(float_filepath)
            self.int_filepaths.append(int_filepath)

    def create_validation_filepaths(self):
        """ Creates a list of observations full filepaths for each
        validation window."""

        for window_index, window in enumerate(self.validation_windows):
            float_filepath, int_filepath = self.make_validation_filepaths(
                                            window.chrom, window_index)
            self.validation_float_filepaths.append(float_filepath)
            self.validation_int_filepaths.append(int_filepath)

    def make_seq_cells(self, chromosome, start, end):
        if self.use_dinucleotide:
            return chromosome.seq[start:end]

    # XXX: Dead code. Observations.save (which calls this function) no longer
    # called. May be used when --observations is re enabled for caching
    # observation files
    # XXX: Determine if posterior will also use a temporary file path as well
    def write_tab_file(self, float_filelist, int_filelist, float_tabfile):
        print_filepaths_custom = partial(self.print_filepaths,
                                         float_filelist, int_filelist,
                                         temp=(self.identify or self.train))

        float_tabwriter = ListWriter(float_tabfile)
        float_tabwriter.writerow(FLOAT_TAB_FIELDNAMES)

        for window_index, window in enumerate(self.windows):
            world, chrom, start, end = window
            float_filepath, int_filepath = \
                print_filepaths_custom(chrom, window_index)

            row = [float_filepath, str(window_index), chrom, str(start),
                   str(end)]
            float_tabwriter.writerow(row)
            print(" %s (%d, %d)" % (float_filepath, start, end), file=sys.stderr)

    def open_writable_or_dummy(self, filepath):
        if not filepath or (not self.clobber and filepath.exists()):
            return closing(StringIO())  # dummy output
        else:
            return open(filepath, "w")

    # XXX: Dead code. Observations.save no longer called. May be used when
    # --observations is re enabled for caching observation files
    def save(self):
        open_writable = self.open_writable_or_dummy

        with open_writable(self.float_filelistpath) as float_filelist:
            with open_writable(self.int_filelistpath) as int_filelist:
                with open_writable(self.float_tabfilepath) as float_tabfile:
                    self.write_tab_file(float_filelist, int_filelist, float_tabfile)



"""segway: a way to segment the genome

Segway is a tool for easy pattern discovery and identification in
functional genomics data.
"""

# Copyright 2008-2014 Michael M. Hoffman <michael.hoffman@utoronto.ca>

from __future__ import absolute_import

import sys

if (sys.version_info[0] == 2 and sys.version_info[1] < 7) or \
   (sys.version_info[0] == 3 and sys.version_info[1] < 4):
    print("Segway requires Python version 2.7 or 3.4 or later")
    sys.exit(1)

from setuptools import find_packages, setup

doclines = __doc__.splitlines()
name, short_description = doclines[0].split(": ")
long_description = "\n".join(doclines[2:])

url = "https://%s.hoffmanlab.org" % name.lower()
download_url = "https://pypi.python.org/pypi/%s" % name.lower()

classifiers = ["Natural Language :: English",
               "Development Status :: 5 - Production/Stable",
               "Intended Audience :: Science/Research",
               "License :: OSI Approved :: GNU General Public License v2 "
               "(GPLv2)",
               "Topic :: Scientific/Engineering :: Bio-Informatics",
               "Operating System :: Unix",
               "Programming Language :: Python",
               "Programming Language :: Python :: 2.7",
               "Programming Language :: Python :: 3"]

entry_points = """
[console_scripts]
segway = segway.run:main
segway-task = segway.task:main
segway-layer = segway.layer:main
segway-winner = segway.winner:main
"""

# XXX: warn: make sure you have LDFLAGS unset if you are building numpy

# need optbuild>0.1.11 for OptionBuilder_ShortOptWithEquals
# need tables>2.04 (>=r3761) because there is a CArray fill bug until then
# genomedata>=1.4.2 for both Python 2 and 3 support
# optplus>=0.2 for both Python 2 and 3 support

setup_requires = ["setuptools_scm"] # source control management packaging
install_requires = ["genomedata>=1.4.2", "autolog>=0.2.0",
                    "textinput>=0.2.0", "optbuild>=0.2.0",
                    "optplus>=0.2.0", "tables>2.0.4", "numpy", "path.py>=11",
                    "colorbrewer>=0.2.0", "drmaa>=0.4a3", "six"]


def main():
    setup(name=name,
          use_scm_version=True,
          description=short_description,
          author="Michael Hoffman",
          author_email="michael.hoffman@utoronto.ca",
          url=url,
          download_url=download_url,
          classifiers=classifiers,
          long_description=long_description,
          setup_requires=setup_requires,
          install_requires=install_requires,
          zip_safe=False, # XXX: change back, this is just for better tracebacks
          packages=find_packages("."),
          include_package_data=True,
          entry_points=entry_points
          )

if __name__ == "__main__":
    main()



from __future__ import absolute_import
from os import remove
from os.path import isfile, sep
from tempfile import gettempdir
import unittest

from numpy import allclose, array, empty

from segway.observations import (merge_windows, VirtualEvidenceWarning,
                                 get_downsampled_virtual_evidence_data_and_presence)
from segway.task import prepare_gmtk_observations, prepare_virtual_evidence
from segway._util import EXT_INT, EXT_FLOAT, EXT_VIRTUAL_EVIDENCE


class TestObservations(unittest.TestCase):

    def test_merge_windows(self):
        self.assertEqual(
            merge_windows([(0, 1)]),
            [(0, 1)])

        self.assertEqual(
            merge_windows([(0, 100), (100, 200)]),
            [(0, 200)])

        self.assertEqual(
            merge_windows([(0, 100), (101, 200)]),
            [(0, 100), (101, 200)])

        self.assertEqual(
            merge_windows([(0, 100), (50, 125), (99, 150), (200, 201)]),
            [(0, 150), (200, 201)])


class TestTask(unittest.TestCase):

    def test_prepare_observations(self):
        gmtk_args = ["baz", "-of1", "foo", "-of2", "bar", "-cppCommandOptions","-DVIRTUAL_EVIDENCE_LIST_FILENAME=VE_PLACEHOLDER"]

        [float_obs_filename, int_obs_filename, virtual_evidence_filename,
         float_obs_list_filename, int_obs_list_filename, virtual_evidence_list_filename] = \
            prepare_gmtk_observations(gmtk_args, "chr1", 0, 8000, empty(0), 1,
                                      None)

        current_tmp_dir = gettempdir()
        # Test proper naming of the observation files
        proper_obs_filename_prefix = current_tmp_dir + sep + "chr1.0.8000"
        self.assertTrue(float_obs_filename.startswith(
                            proper_obs_filename_prefix))
        self.assertTrue(float_obs_filename.endswith(EXT_FLOAT))

        self.assertTrue(int_obs_filename.startswith(
                            proper_obs_filename_prefix))
        self.assertTrue(int_obs_filename.endswith(EXT_INT))

        self.assertTrue(virtual_evidence_filename.startswith(
                            proper_obs_filename_prefix))
        self.assertTrue(virtual_evidence_filename.endswith(EXT_VIRTUAL_EVIDENCE))

        # Test filenames exist in the given environment's tempdir
        self.assertTrue(float_obs_list_filename.startswith(current_tmp_dir))
        self.assertTrue(int_obs_list_filename.startswith(current_tmp_dir))
        self.assertTrue(virtual_evidence_list_filename.startswith(current_tmp_dir))
        self.assertTrue(float_obs_filename.startswith(current_tmp_dir))
        self.assertTrue(int_obs_filename.startswith(current_tmp_dir))
        self.assertTrue(virtual_evidence_filename.startswith(current_tmp_dir))

        # Check if the observation files exist
        self.assertTrue(isfile(float_obs_filename))
        self.assertTrue(isfile(int_obs_filename))
        self.assertTrue(isfile(virtual_evidence_filename))

        # Check the first line in the list contains their respective
        # observation file (and type)
        with open(float_obs_list_filename) as float_list_file:
            line = float_list_file.readlines()[0].strip()
            self.assertTrue(line.startswith(proper_obs_filename_prefix))
            self.assertTrue(line.endswith(EXT_FLOAT))

        with open(int_obs_list_filename) as int_list_file:
            line = int_list_file.readlines()[0].strip()
            self.assertTrue(line.startswith(proper_obs_filename_prefix))
            self.assertTrue(line.endswith(EXT_INT))

        with open(virtual_evidence_list_filename) as virtual_evidence_list_file:
            line = virtual_evidence_list_file.readlines()[0].strip()
            self.assertTrue(line.startswith(proper_obs_filename_prefix))
            self.assertTrue(line.endswith(EXT_VIRTUAL_EVIDENCE))
        # Check gmtk args were modified for the temp list files
        # -of1 option
        self.assertEqual(gmtk_args[2], float_obs_list_filename)
        # -of2 option
        self.assertEqual(gmtk_args[4], int_obs_list_filename)
        # virtual evidence cpp directive
        self.assertEqual(gmtk_args[6], "-DVIRTUAL_EVIDENCE_LIST_FILENAME=%s" % virtual_evidence_list_filename)

        remove(float_obs_list_filename)
        remove(int_obs_list_filename)
        remove(virtual_evidence_list_filename)
        remove(int_obs_filename)
        remove(float_obs_filename)
        remove(virtual_evidence_filename)


class TestVirtualEvidence(unittest.TestCase):
    def test_proper_downsampling(self):
        virtual_evidence_coords = "[(0, 4), (0, 2), (2, 4)]"
        virtual_evidence_priors = "[{0: 0.1}, {1: 0.5}, {2: 0.7}]"

        raw = prepare_virtual_evidence(1, 0, 5, 4,
                                       virtual_evidence_coords,
                                       virtual_evidence_priors)
        expected = array([[0.1, 0.5, 0.2, 0.2], [0.1, 0.5, 0.2, 0.2],
                          [0.1, 0.1, 0.7, 0.1], [0.1, 0.1, 0.7, 0.1],
                          [0, 0, 0, 0]])
        self.assertTrue(allclose(raw, expected))

        downsampled, presence = get_downsampled_virtual_evidence_data_and_presence(
            raw, resolution=5, num_labels=4)
        self.assertTrue(allclose(downsampled, array([0.1, 0.3, 0.45, 0.15])))
        print(presence)

    def test_overlapping_labels(self):
        virtual_evidence_priors = "[{0: 0.2}, {0: 0.7}]"
        virtual_evidence_coords = "[(0, 100), (50, 150)]"

        with self.assertRaises(ValueError):
            prepare_virtual_evidence(1, 0, 800, 4,
                                     virtual_evidence_coords,
                                     virtual_evidence_priors)

    def test_zero_labels(self):
        virtual_evidence_coords = "[(0, 4), (4, 5)]"
        virtual_evidence_priors = "[{1: 0.8}, {0: 0}]"

        expected = prepare_virtual_evidence(1, 0, 5, 3,
                                           virtual_evidence_coords,
                                           virtual_evidence_priors)
        self.assertTrue((expected[4] == array([0, 0.5, 0.5])).all())

    def test_negative_priors(self):
        virtual_evidence_coords = "[(0, 4), (2, 5)]"
        virtual_evidence_priors = "[{1: -0.8}, {0: 0.7}]"

        with self.assertRaises(ValueError):
            prepare_virtual_evidence(1, 0, 10, 4,
                                     virtual_evidence_coords,
                                     virtual_evidence_priors)

    def test_priors_over_one(self):
        virtual_evidence_coords = "[(0, 4), (2, 5)]"
        virtual_evidence_priors = "[{1: 0.8}, {0: 0.7}]"

        # Only python 3 supports assetWarns, passes if there is an
        # AttributeError in python 2
        # Should be dropped once python 2 support is
        try:
            with self.assertWarns(VirtualEvidenceWarning):
                prepare_virtual_evidence(1, 0, 10, 4,
                                         virtual_evidence_coords,
                                         virtual_evidence_priors)
        except AttributeError:
            pass


if __name__ == "__main__":
    unittest.main()
